-- Create ACE Addon
TRCLOOT = TRCLOOT or LibStub("AceAddon-3.0"):NewAddon("TRCLOOT", "AceConsole-3.0")
local t = TRCLOOT



t.ADDON_NAME = "RavenLoot"



--[[ STATWEIGHTS ]]--
t.defaultStatWeights = {

    Intellect = 1,
    Agility = 1,
    Strength = 1,

    Haste = 1,
    Mastery = 1,
    Crit = 1,
    Versatility = 1,

    Stamina = 0,
    Armor = 0,

    Leech = 0,
    Avoidance = 0,
    Speed = 0
}


-- Returns the saved var, or default value if not available
t.getVar = function(varName) 
    if (not t.svc) or (t.svc[varName] == nil) then
        t.svc = t.svc or {}
        t.svc[varName] = t.defaultVars[varName]
    end


    return t.svc[varName]

end

-- Set a saved var
t.setVar = function(varName, value)
    t.svc[varName] = value
end

-- Shallow copy a table
t.tableCopy = function(l1)
   local l2 = {}
   for k, v in pairs(l1) do
       l2[k] = v
   end
   
   return l2
end

    

-- TODO: Possible bug: Specs share same sw table?
local firstSet = GetEquipmentSetInfo(1)
t.defaultVars = {
    font = "Friz Quadrata TT",
    fontPath = "Fonts\\FRIZQT__.TTF",
    showRealmNames = false,
    showMoveMe = true,
    reportAllEquipable = false,
    specs = {{enable=false, ms=false, equipmentSet=firstSet, sw=t.tableCopy(t.defaultStatWeights)}, 
             {enable=false, ms=false, equipmentSet=firstSet, sw=t.tableCopy(t.defaultStatWeights)}, 
             {enable=false, ms=false, equipmentSet=firstSet, sw=t.tableCopy(t.defaultStatWeights)}, 
             {enable=false, ms=false, equipmentSet=firstSet, sw=t.tableCopy(t.defaultStatWeights)}, 
         }
}



-- UI CONFIG
t.ui = {
    ANCHOR = "TOPLEFT",

    MAIN_WINDOW_WIDTH = 340,
    ITEM_HEIGHT = 60,
    ITEM_INNER_MARGIN = 5,
    ITEM_SPACING = 5,

    SIDE_WINDOW_WIDTH = 260,
    PLAYER_HEIGHT = 18,
    PLAYER_SPACING = 4,
    PLAYER_INNER_MARGIN = 4,


}


-- Stat mapping from WoW stat names to stat weight indices.
t.stm = {
    ITEM_MOD_INTELLECT_SHORT = "Intellect",
    ITEM_MOD_AGILITY_SHORT = "Agility",
    ITEM_MOD_STRENGTH_SHORT = "Strength",

    ITEM_MOD_HASTE_RATING_SHORT = "Haste",
    ITEM_MOD_MASTERY_RATING_SHORT = "Mastery",
    ITEM_MOD_CRIT_RATING_SHORT = "Crit",
    ITEM_MOD_VERSATILITY = "Versatility",

    ITEM_MOD_STAMINA_SHORT = "Stamina",
    RESISTANCE0_NAME = "Armor",

    ITEM_MOD_CR_AVOIDANCE_SHORT = "Avoidance",
    ITEM_MOD_CR_SPEED_SHORT = "Speed",
    ITEM_MOD_CR_LIFESTEAL_SHORT = "Leech"


}

t.reverseStm = {}
for fullName, shortName in pairs(t.stm) do
   t.reverseStm[_G[fullName]] = fullName
end

-- Slot mapping from inventory slot names to inventory numbers
t.slm = {
    INVTYPE_HEAD = 1,
    INVTYPE_NECK = 2,
    INVTYPE_SHOULDER = 3,
    INVTYPE_CHEST = 5,
    INVTYPE_ROBE = 5,
    INVTYPE_WAIST = 6,
    INVTYPE_LEGS = 7,
    INVTYPE_FEET = 8,
    INVTYPE_WRIST = 9,
    INVTYPE_HAND = 10,
    INVTYPE_FINGER = 11, 
    INVTYPE_TRINKET = 13, 
    INVTYPE_CLOAK = 15,
    INVTYPE_SHIELD = 17,
}

t.primaryStats = {
   ITEM_MOD_INTELLECT_SHORT=true,
   ITEM_MOD_STRENGTH_SHORT=true,
   ITEM_MOD_AGILITY_SHORT=true,
}


t.armorTypes = {
    Cloth = {PRIEST=true, MAGE=true, WARLOCK=true},
    Leather = {ROGUE=true, MONK=true, DRUID=true, DEMONHUNTER=true},
    Mail = {HUNTER=true, SHAMAN=true},
    Plate = {WARRIOR=true, PALADIN=true, DEATHKNIGHT=true}
}


t.upgradeIDs = {
    [529] = 531,
    [530] = 531,
    [531] = 531,
}

-- Init code
t.libS = t.libS or LibStub:GetLibrary("AceSerializer-3.0")

-- Constants, mostly for communication
t.VERSION_NUMBER = 0.5
t.ADDON_MESSAGE_PREFIX = "TRCLOOT_MPREF"
t.ADDON_MESSAGE_MAX_LENGTH = 253 - t.ADDON_MESSAGE_PREFIX:len()
t.MSG_NOT_DONE = "0"
t.MSG_DONE = "1"
t.OFFER = "1"
t.REPORT = "2"
t.DECLINE =  "3"
t.VERSION_REQUEST ="4"
t.VERSION_RESPONSE = "5"
t.VERSION_TIMEOUT = 1500 --ms

t.MSG_PREFIX = "|cff60d4d4RavenLoot: |r"

-- Very ugly hack. If the score of the currently equiped item ends up being 0
-- (which is very unlikely), assign this number as percent upgrade.
-- TODO: Fix :>
t.INFINITE_UPGRADE = 2000000000

-- Register addon prefix 
RegisterAddonMessagePrefix(t.ADDON_MESSAGE_PREFIX)


t.offerLoot = function(link, selfOnly, origin)
    local t = TRCLOOT

    local data = {
        type = t.OFFER,
        hyperlink = t.getMaxUpgrade(link),
        selfOnly = selfOnly,
        origin = origin
    }
    local serialized = t.libS:Serialize(data)

    if IsInGroup() and not selfOnly then
        t.sendAddonMessage(serialized, "RAID")
    else
        local shortName = GetUnitName("player")
        local realmName = GetRealmName():gsub("%s+", "")
        local name = shortName.."-"..realmName

        t.sendAddonMessage(serialized, "WHISPER", name)
    end


end

-- Define usage function
--t.printUsage = t.printUsage or function()
t.printUsage = function()
    print("|cff60d4d4Raven Loot|r command list: ")
    print("|cff60d4d4/rloot|r |cffffff33offer|r |cfffcffb0<item link>|r: Offer an item to the group.")
    print("|cff60d4d4/rloot|r |cffffff33compare|r |cfffcffb0<item link>|r: Show an item just for yourself.")
    print("|cff60d4d4/rloot|r |cffffff33versions|r: List raid members that don't have the addon or are using an outdated version.")
    print("|cff60d4d4/rloot|r |cffffff33move|r: Toggle the movable anchor.")
    print("|cff60d4d4/rloot|r |cffffff33config|r: Open the configuration panel.")
end


-- Register slash command and parse it
t:RegisterChatCommand("rloot", function(msg, editbox)

    local t = TRCLOOT

    -- Split on space
    local words = {}
    for word in msg:gmatch("[^ ]+") do 
        table.insert(words, word)
    end

    -- if offer command
    if words[1] == "offer" then
        table.remove(words, 1)
        local hyperlink = table.concat(words, " ")
        -- If no link, print usage
        if hyperlink == "" then 
            t.printUsage()
        else    
            t.offerLoot(hyperlink, false)
        end 
    elseif words[1] == "compare" then
        table.remove(words, 1)
        local hyperlink = table.concat(words, " ")
        -- If no link, print usage
        if hyperlink == "" then 
            t.printUsage()
        else    
            t.offerLoot(hyperlink, true)
        end 
    elseif words[1] == "versions" then
        t.checkVersions()
    elseif words[1] == "move" then
        t.toggleMoveMe()
    elseif words[1] == "config" then
        InterfaceOptionsFrame_OpenToCategory(t.ADDON_NAME)
        InterfaceOptionsFrame_OpenToCategory(t.ADDON_NAME)
    else
        t.printUsage()
    end

end)


-- Show the MoveMe 
t.toggleMoveMe = function()
    t.setVar("showMoveMe", not t.getVar("showMoveMe"))
    t.drawReports()
end


-- Reports players that are not using the addon or are using
-- an outdated version.
t.checkVersions = function()

    print(t.MSG_PREFIX.."Sending out version requests.")
    t.versionInfo = t.versionInfo or {}
    t.versionInfo = table.wipe(t.versionInfo)

    -- Broadcast a version request
    local data = {
        type = t.VERSION_REQUEST,
    }
    local serialized = t.libS:Serialize(data)
    t.sendAddonMessage(serialized, "RAID")

    -- After a while, show results
    C_Timer.After(t.VERSION_TIMEOUT/1000, function() 

        -- Print players in raid that didn't respond (no addon?)
        local missing = {}
        for n=1,40 do
            local playerInfo = {GetRaidRosterInfo(n)}
            if playerInfo[1] then -- if this member exists
                local name = playerInfo[1]
                local shortName = Ambiguate(name, "short")


                if not t.versionInfo[shortName] then -- if no response
                    table.insert(missing, name)
                end
            end
        end

        if #missing > 0 then
            print(t.MSG_PREFIX.."No response from the following players: ")
            for n=1, #missing do
                print("      "..missing[n])
            end
        end


        local highestVersion = t.VERSION_NUMBER
        local outdated = {}

        -- Loop over responses
        for player, version in pairs(t.versionInfo) do
            if version > highestVersion then
                highestVersion = version
            elseif version < t.VERSION_NUMBER then
                outdated[player] = version
            end
        end

        if not (next(outdated) == nil) then -- print outdated players
            print(t.MSG_PREFIX.."Players running an older version than you:")
            for player, version in pairs(outdated) do
                print("      "..player.." ("..version..")")
            end
        end

        if not (highestVersion == t.VERSION_NUMBER) then -- Warn if you are outdated
            print(t.MSG_PREFIX.."You are running an outdated version! You are using "..t.VERSION_NUMBER..", the newest found is "..highestVersion..".")
        elseif next(outdated) == nil then -- All using same versions
            print(t.MSG_PREFIX.."Everyone that responded is using the same version ("..t.VERSION_NUMBER..").")
        end

    end)

end

-- Responds to a version request
t.respondToVersionRequest = function(sender)
    local data = {
        type = t.VERSION_RESPONSE,
        versionNumber = t.VERSION_NUMBER
    }
    local serialized = t.libS:Serialize(data)
    t.sendAddonMessage(serialized, "WHISPER", sender)    
end




-- Returns class colors or the fallback if not possible
t.getClassColor = function(fullName, fallback)

    local shortName = Ambiguate(fullName, "none")
    local unitClass = {UnitClass(shortName)}

    if unitClass[2] then
        return RAID_CLASS_COLORS[unitClass[2]]
    end
    return fallback

end


t.getMaxUpgrade = function(item)

    if not item then 
        return item
    end

    -- Split on : 
    local l = {}
    for t in item:gmatch("([^:]*):?") do
        table.insert(l, t)
    end

    -- Set last entry to nil, it is an empty string
    l[#l] = nil

    -- Get current upgradeID
    local upgradeID = tonumber(l[17])

    -- If no upgradeID, return item as it was
    if (not upgradeID) or (upgradeID == "") or (upgradeID == 0) then
        return item
    end

    -- Get the fully upgraded version from the upgradeIDs table
    local newUpgradeID = t.upgradeIDs[upgradeID]


    -- If there was a corresponding maxed out upgrade id, use it
    if newUpgradeID then
        -- Insert the new upgradeID
        l[17] = newUpgradeID

        -- Join and return 
        return table.concat(l, ":")
    else
        -- If no known max upgrade id, just use the item as is.
        return item
    end

end



-- Calculates weighted stat sum for a given item link
t.calculateItemScore = function(item, statWeights, primaryStat, gemStat, gemStatVal)
    local t = TRCLOOT


    -- Get the stats


    local stats = t.getItemStats(item, primaryStat, gemStat, gemStatVal)

    if stats == nil then
        print("Stats not loaded in calc i score")
        return nil
    end


    if next(stats) == nil and (not IsArtifactRelicItem(item)) then
        print("Watch out, no stats found on item "..item..". If this is incorrect, please report it as a bug.")
    end


    -- Calculate the weighted sum
    local score = 0

    for stat, val in pairs(stats) do
        -- TODO: MAKE SURE IT IS LOADED AND ADD A CALLBACK
        --local statWeights = t.getVar("specs")[spec]["sw"]
        local statWeight = statWeights[t.stm[stat]] or 0
        local statScore = statWeight * val
        score = score + statScore
    end

    return score
end

-- Function to extract itemID from itemLink. 
t.getItemID = function(itemLink)
    local itemInfo = {}
    for token in string.gmatch(itemLink, '([^:]+)') do
        table.insert(itemInfo, token)
    end
    return itemInfo[2]
end



t.storeItem = function(item, origin)
    -- Store the item
    local itemID = t.getItemID(item)
    t.reports = t.reports or {}
    t.reports[origin] = t.reports[origin] or {}
    t.reports[origin][itemID] = t.reports[origin][itemID] or {}
    t.reports[origin][itemID].item = item
end


-- Compares an item with currently equiped item 
-- and broadcast the score
t.calculateUpgradeScore = function(item, origin, spec, selfOnly)
    local t = TRCLOOT

    -- Find slot name
    local info = {GetItemInfo(item)}

    -- See if Item info available
    if next(info) == nil then
        -- register callback
        t.itemInfoQueue = t.itemInfoQueue or {}
        table.insert(t.itemInfoQueue, {item=item, origin=origin, spec=spec, callback=t.calculateUpgradeScoreCallback, selfOnly=selfOnly})
    else
        t.calculateUpgradeScoreCallback(item, origin, spec, selfOnly)
    end


end

t.calculateUpgradeScoreCallback = function(item, origin, spec, selfOnly)

    if spec == GetSpecialization() then
        t.calculateUpgradeScoreCurrentSpec(item, origin, spec, selfOnly)
    else
        t.calculateUpgradeScoreOtherSpec(item, origin, spec, selfOnly)
    end

end

t.getItemFromSet = function(itemsLocations, equipSlot) 

    local itemLocation = itemsLocations[equipSlot]
    local player, bank, bags, slot, bag, bagNum = EquipmentManager_UnpackLocation(itemLocation)

    local equipedItem
    if player and (not bags) then
        return GetInventoryItemLink("player", bag)
    elseif player and bags then
        return GetContainerItemLink(bagNum, bag)
    else
        return nil
    end
end



t.calculateUpgradeScoreOtherSpec = function(item, origin, spec, selfOnly)
    local t = TRCLOOT

    -- Relics currently not supported for IS
    if IsArtifactRelicItem(item) then
        return 
    end

    local info = {GetItemInfo(item)}
    if not t.isEquipable(item, info) then
        return
    end


    local specInfo = t.getVar("specs")[spec]
    local setName = specInfo["equipmentSet"]

    local equipLoc = info[9]

    if equipLoc == "" then
        return
    end

    local equipSlot = t.slm[equipLoc]


    local itemsLocations = GetEquipmentSetLocations(setName)
    local equipedItem = t.getItemFromSet(itemsLocations, equipSlot)

    if not equipedItem then
        local specName = select(2, GetSpecializationInfo(spec))
        print("Cannot find item to compare with "..item.." for spec "..specName.." from equipment set "..setName..".")
        return
    end

    -- TODO: Relics
    local equipedItem2
    if (equipLoc == "INVTYPE_FINGER") or (equipLoc == "INVTYPE_TRINKET") then
        equipedItem2 = t.getItemFromSet(itemsLocations, equipSlot+1)
        equipedItem2 = t.getMaxUpgrade(equipedItem2)
    end

    -- If only 1 ring/trinket, set it on slot 1
    if (not equipedItem) and equipedItem2 then
        equipedItem = equipedItem2
        equipedItem2 = nil
    end


    -- If at least 1 equiped
    if equipedItem then
        local info = {GetItemInfo(equipedItem)}

        -- See if Item info available
        if next(info) == nil then
            -- register callback
            t.itemInfoQueue = t.itemInfoQueue or {}
            table.insert(t.itemInfoQueue, {item=item, equipedItem=equipedItem, equipedItem2=equipedItem2, origin=origin, spec=spec, callback=t.calculateUpgradeScoreCallback2})
        else
            t.calculateUpgradeScoreCallback2(item, equipedItem, equipedItem2, origin, spec, selfOnly)
        end
    else
        t.calculateUpgradeScoreCallback2(item, equipedItem, equipedItem2, origin, spec, selfOnly)
    end

end




t.calculateUpgradeScoreCurrentSpec = function(item, origin, spec, selfOnly)
    local info = {GetItemInfo(item)}

    -- If can't equip this (or wrong armor type) decline it.
    if not t.isEquipable(item, info) then
        local data = {
            type = DECLINE,
            origin = origin,
            item = item
        }
        local serialized = t.libS:Serialize(data)
        t.sendAddonMessage(serialized, "RAID")
        return
    end

    local equipedItem
    local equipedItem2

    -- If  relic
    if IsArtifactRelicItem(item) then

        equipedItem = t.equipedRelic
        t.equipedRelic = nil

        equipedItem2 = t.equipedRelic2
        t.equipedRelic = nil

    else -- If not relic
        local equipSlotNumber = t.slm[info[9]]

        -- Find currently equiped item in that slot
        equipedItem = GetInventoryItemLink("player", equipSlotNumber)
        equipedItem = t.getMaxUpgrade(equipedItem)

        if info[9] == "INVTYPE_FINGER" or info[9] == "INVTYPE_TRINKET" then
            equipedItem2 = GetInventoryItemLink("player", equipSlotNumber+1)
            equipedItem2 = t.getMaxUpgrade(equipedItem2)
        end

        -- If only 1 ring/trinket, set it on slot 1
        if (not equipedItem) and equipedItem2 then
            equipedItem = equipedItem2
            equipedItem2 = nil
        end
    end

    -- If at least equiped
    if equipedItem then
        local info = {GetItemInfo(equipedItem)}

        -- See if Item info available
        if next(info) == nil then
            -- register callback
            t.itemInfoQueue = t.itemInfoQueue or {}
            table.insert(t.itemInfoQueue, {item=item, equipedItem=equipedItem, equipedItem2=equipedItem2, origin=origin, spec=spec, callback=t.calculateUpgradeScoreCallback2, selfOnly=selfOnly})
        else
            t.calculateUpgradeScoreCallback2(item, equipedItem, equipedItem2, origin, spec, selfOnly)
        end
    else
        t.calculateUpgradeScoreCallback2(item, equipedItem, equipedItem2, origin, spec, selfOnly)
    end



end


t.setItemToTooltip = function(item, d, callback)
   local tooltipName = "RavenCouncil_ItemScanningTooltip"
   t.scanTooltip = t.scanTooltip or CreateFrame("GameTooltip", tooltipName, UIParent, "GameTooltipTemplate")

   t.scanQueue = t.scanQueue or {}
  
   table.insert(t.scanQueue, {item=item, d=d, callback=callback})

   t.scanNextInQueue()
end


t.scanNextInQueue = function()
    local tooltipName = "RavenCouncil_ItemScanningTooltip"

    if t.currentlyScanning == nil then
        t.currentlyScanning = false
    end

    if (not t.currentlyScanning) and ((#t.scanQueue) > 0) then
        t.currentlyScanning = true

        local data = table.remove(t.scanQueue, 1)

        t.scanTooltip:ClearLines()
        t.scanTooltip:SetOwner(UIParent, "ANCHOR_NONE")
        t.scanTooltip:SetScript("OnUpdate", function()
            if (_G[tooltipName.."TextLeft1"]:GetText() ~= "Retrieving item information")  then
                t.scanTooltip:SetScript("OnUpdate", nil)
                    data.callback(data.d)
                    t.currentlyScanning = false
                    t.scanNextInQueue()
            end
        end)
        t.scanTooltip:SetHyperlink(data.item)
    end
end

t.calculateUpgradeScoreCallback2 = function(item, equipedItem, equipedItem2,  origin, spec, selfOnly)
    local t = TRCLOOT
    
    local d = {}
    d.specs = t.getVar("specs")
    d.statWeights = d.specs[spec]["sw"]
    d.primaryStat = d.specs[spec]["primary"]
    d.gemStat = d.specs[spec]["gemStat"]
    d.gemStatVal = d.specs[spec]["gemStatVal"]
    d.item = item
    d.equipedItem = equipedItem
    d.equipedItem2 = equipedItem2
    d.origin = origin
    d.spec = spec
    d.selfOnly = selfOnly

    -- Get the item scores
    d.newScore = 0
    if d.item then 
        d.item = t.getMaxUpgrade(d.item)
        t.setItemToTooltip(d.item, d, t.cCB3)
    else
        t.cCB3(d)
    end


end


t.cCB3 = function(d)

    if d.item then
        d.newScore = t.calculateItemScore(d.item, d.statWeights, d.primaryStat, d.gemStat, d.gemStatVal)
    end

    d.equipedScore = 0
    if d.equipedItem then
        d.equipedItem = t.getMaxUpgrade(d.equipedItem)
        t.setItemToTooltip(d.equipedItem, d, t.cCB4)
    else

        t.cCB4(d)
    end
end

t.cCB4 = function(d)

    if d.equipedItem then
        d.equipedScore = t.calculateItemScore(d.equipedItem, d.statWeights, d.primaryStat, d.gemStat, d.gemStatVal)
    end


    d.equipedScore2 = nil
    if d.equipedItem2 then
        d.equipedItem2 = t.getMaxUpgrade(d.equipedItem2)
        t.setItemToTooltip(d.equipedItem2, d, t.cCB5)
    else
        t.cCB5(d)
    end
end

t.cCB5 = function(d)

    if d.equipedItem2 then
        d.equipedScore2 = t.calculateItemScore(d.equipedItem2, d.statWeights, d.primaryStat, d.gemStat, d.gemStatVal)
    end


    t.cCB7(d)

end

t.cCB7 = function(d)
    local item = d.item
    local newScore = d.newScore

    local equipedItem = d.equipedItem
    local equipedScore =  d.equipedScore 

    local equipedItem2 = d.equipedItem2
    local equipedScore2 = d.equipedScore2

    local origin = d.origin
    local spec = d.spec
    local primaryStat = d.primaryStat


    -- If relic
    local relicIlvlUpgrade
    if IsArtifactRelicItem(item) then
        local itemIlvl = C_ArtifactUI.GetItemLevelIncreaseProvidedByRelic(item)
        if equipedItem then
             relicIlvlUpgrade =  itemIlvl - C_ArtifactUI.GetItemLevelIncreaseProvidedByRelic(equipedItem) 
        end
        if equipedItem2 then
             local equipedItem2Upgrade =  itemIlvl - C_ArtifactUI.GetItemLevelIncreaseProvidedByRelic(equipedItem2)
             if equipedItem2Upgrade > relicIlvlUpgrade then
                 relicIlvlUpgrade = equipedItem2Upgrade
             end
        end
    end


    local lowestScore = equipedScore
    if equipedItem2 and (equipedScore >  equipedScore2) then 
        lowestScore = equipedScore2
    end


    -- Calculate percent upgrade
    local percentUpgrade
    if lowestScore == 0 then
        percentUpgrade = t.INFINITE_UPGRADE
    else
        percentUpgrade = (newScore - lowestScore) / lowestScore * 100
    end


    if relicIlvlUpgrade then
        percentUpgrade = relicIlvlUpgrade
    end

    local className, classID = select(2, UnitClass("player"))

    local currentSpecInfo = t.getVar("specs")[spec]

    -- Serialize and print
    local data = {
        type = t.REPORT,
        equipedItem = equipedItem,
        equipedScore = equipedScore,
        equipedItem2 = equipedItem2,
        equipedScore2 = equipedScore2,
        newItem = item,
        newScore = newScore,
        origin = origin,
        percentUpgrade = percentUpgrade,
        primaryStat = primaryStat,
        gemStat = currentSpecInfo["gemStat"],
        gemStatVal = currentSpecInfo["gemStatVal"],
        statWeights = currentSpecInfo["sw"],
        className = className, 
        classID = classID,
        spec = spec,
        ms = currentSpecInfo["ms"],
        relic = (relicIlvlUpgrade ~= nil)
    }
    local serialized = t.libS:Serialize(data)    
    if IsInGroup() and not d.selfOnly then
        t.sendAddonMessage(serialized, "RAID")
    else
        local shortName = GetUnitName("player")
        local realmName = GetRealmName():gsub("%s+", "")
        local name = shortName.."-"..realmName

        t.sendAddonMessage(serialized, "WHISPER", name)

    end

    function tprint (tbl, indent)
        if not indent then indent = 0 end
        for k, v in pairs(tbl) do
            formatting = string.rep("  ", indent) .. k .. ": "
            if type(v) == "table" then
                print(formatting)
                tprint(v, indent+1)
            elseif type(v) == 'boolean' then
                print(formatting .. tostring(v))      
            else
                print(formatting .. v)
            end
        end
    end
    
    --tprint(data)


end

t.isEquipable = function(item, info) 
    local t = TRCLOOT




    -- Scan for Class restrictions
    local tooltipName = "RavenCouncil_ItemScanningTooltip"
    t.scanTooltip = t.scanTooltip or CreateFrame("GameTooltip", tooltipName, UIParent, "GameTooltipTemplate")
    t.scanTooltip:SetOwner(UIParent, "ANCHOR_NONE")


    local usable = true


    t.scanTooltip:ClearLines()
    t.scanTooltip:SetHyperlink(item)

    local cLine 
    for n=1, t.scanTooltip:NumLines() do
        local line = _G[tooltipName.."TextLeft"..n]:GetText()..(_G[tooltipName.."TextRight"..n]:GetText() or "")
        if string.find(line, "Classes: ") then
            cLine = line
            usable = false
            local classes = line:sub(line:find(":")+2)
            for class in gmatch(classes, "([a-zA-Z][a-zA-Z ]*)") do
                if UnitClass("player") == class then
                    usable = true
                end
            end

            break
        end 
    end

    if not usable then
        return false or t.getVar("reportAllEquipable")
    end

    -- If relic
    if IsArtifactRelicItem(item) then

        local itemID = t.getItemID(item)
        --local relicType = C_ArtifactUI.GetRelicInfoByItemID(itemID)
        --local numSlots = C_ArtifactUI.GetEquippedArtifactNumRelicSlots()
        --

        local equipable = false
        local relicSlotCount = C_ArtifactUI.GetEquippedArtifactNumRelicSlots()

        -- If no artifact equiped
        if not relicSlotCount then
            return
        end

        for n=1,  relicSlotCount do
            if C_ArtifactUI.CanApplyRelicItemIDToEquippedArtifactSlot(itemID, n) then
                local equipedRelic = select(4, C_ArtifactUI.GetEquippedArtifactRelicInfo(n))
                if t.equipedRelic then
                    t.equipedRelic2 = equipedRelic
                else
                    t.equipedRelic = equipedRelic
                end
                

                equipable = true
            end
        end
        return equipable or t.getVar("reportAllEquipable")
    end





    -- Check for armor type
    local info = {GetItemInfo(item)}

    local itemType = info[6]
    local itemSubType = info[7]
    local itemSlot = info[9]

    -- If relic


    if not (itemType == "Armor") then -- later add weapon support maybe?
        return false
    end

    -- Report everything else equipable. For debugging.
    if t.getVar("reportAllEquipable") then
        return true
    end

    local className = select(2, UnitClass("player"))
    local classesForArmorType = t.armorTypes[itemSubType]

    -- If cloth/leather/mail/plate
    if classesForArmorType then
        -- If it is of the wrong type, return false
        if (not classesForArmorType[className]) and (itemSubType ~= "Miscellaneous") and (itemSlot ~= "INVTYPE_CLOAK") then
            return false
        end
        return true
        -- If ring / neck / trinket
    elseif itemSubType == "Miscellaneous" then
        return true
    else
        print(t.MSG_PREFIX.."Unknown armor subtype "..itemSubType..". Please report this issue if you believe this type should be recognized.")
        return false
    end
end


-- Serialize data and send it in chunks
t.sendAddonMessage = function(msg, target, player)
    local done = false

    while not done do
        done = msg:len() <= t.ADDON_MESSAGE_MAX_LENGTH
        local chunk = msg:sub(1, t.ADDON_MESSAGE_MAX_LENGTH)
        msg = msg:sub(t.ADDON_MESSAGE_MAX_LENGTH+1)
        local prefix = done and t.MSG_DONE or t.MSG_NOT_DONE
        SendAddonMessage(t.ADDON_MESSAGE_PREFIX, prefix..chunk, target, player)

    end
end

-- Store report response
t.storeReport = function(data)
    local t = TRCLOOT
    t.reports = t.reports or {}


    -- Get item id
    local itemID = t.getItemID(data.newItem)

    -- Store the report
    t.reports[data.origin] = t.reports[data.origin] or {}
    t.reports[data.origin][itemID] = t.reports[data.origin][itemID] or {}
    t.reports[data.origin][itemID].responses = t.reports[data.origin][itemID].responses or {}

    -- The table to insert in
    local l = t.reports[data.origin][itemID].responses


    data.scaledPercentUpgrade = 100+data.percentUpgrade

    local scoreFactors = nil
    if t.scoreFactorFunction then
        scoreFactors = t.scoreFactorFunction()
    end

    if scoreFactors and scoreFactors[data.sender] then
        data.scoreFactor = scoreFactors[data.sender]
        data.scaledPercentUpgrade = scoreFactors[data.sender] * data.scaledPercentUpgrade
    end


    -- Check if entry exists. If so, delete it
    local len = table.getn(l)
    for i=1, len do
        if l[i].sender == data.sender and l[i].spec == data.spec then
            table.remove(l,i)
            len = len - 1
            break
            --l[i] = data
            --return
        end
    end


    -- If the entry doesn't exist, insert it using insertion sort
    local i = 1

    while (i <= len) and (data.scaledPercentUpgrade < l[i].scaledPercentUpgrade) do
        i = i + 1
    end

    if i > len then
        l[i] = data 
    else
        table.insert(l, i, data)
    end






end


t.calculateStatDifferences = function(stats1, stats2, primaryStat, gemStat, gemStatVal)

    --local stats1 = t.getItemStats(item1, primaryStat, gemStat, gemStatVal)
    --local stats2 = t.getItemStats(item2, primaryStat, gemStat, gemStatVal)

    if (stats1 == nil) then
        print ("s1 nil") 
        return nil
    end

    if (stats2 == nil) then
        print("s2 nil")
        return nil
    end


    local diff = {}

    for stat, v1 in pairs(stats1) do
        local v2 = stats2[stats]
        if v2 then
            diff[stat] = v1-v2
        else 
            diff[stat] = v1
        end
    end

    for stat, v2 in pairs(stats2) do
        local v1 = stats1[stat]
        if v1 then
            diff[stat] = v1-v2
        else
            diff[stat] = (diff[stat] or 0) - v2
        end
    end

    return diff
end


t.isMainStat = function(stat)
    return stat == "ITEM_MOD_AGILITY_SHORT" or 
    stat == "ITEM_MOD_STRENGTH_SHORT" or 
    stat == "ITEM_MOD_INTELLECT_SHORT"
end

t.getMainStatItem = function(item)

    if not item then
        return nil
    end

    local stats = ({GetItemStats(item)})[1]
    if not stats then
        print("Could not find stats on item "..item..". This should not happen, oops...")
    end

    if stats["ITEM_MOD_AGILITY_SHORT"] then 
        return "ITEM_MOD_AGILITY_SHORT"
    elseif stats["ITEM_MOD_STRENGTH_SHORT"] then 
        return "ITEM_MOD_STRENGTH_SHORT"
    elseif stats["ITEM_MOD_INTELLECT_SHORT"] then 
        return "ITEM_MOD_INTELLECT_SHORT"
    else
        return nil
    end
end


-- This function attempts to get stats from an item link as if viewed by 
-- a player with primaryStat as the primary stat for their specialization.
t.getItemStats = function(item, primaryStat, gemStat, gemStatVal)
   local res = {}
   local tooltipName = "RavenCouncil_ItemScanningTooltip"
   


   -- If not in cache yet, return nil
   if t.scanTooltip:NumLines() > 0 and _G[tooltipName.."TextLeft1"]:GetText() == "Retrieving item information" then
       return nil
   end
   
   local cLine 
   local primaryCount = 0
   local containsMyPrimary = false
   for n=1, t.scanTooltip:NumLines() do
      local line = _G[tooltipName.."TextLeft"..n]:GetText()..(_G[tooltipName.."TextRight"..n]:GetText() or "")
     -- Remove color codes 
      line = line:match("|c%x%x%x%x%x%x%x%x([^|]+)|r") or line
     -- Extract the stat name and value
      local val, stat = line:match("^([+-]?[0-9,\\.]*) ([a-zA-Z ]+)")
      if val and stat then
          -- Remove the , or . incase the value is larger than 999
          val = val:gsub("[,\\.]", "")
          val = tonumber(val)
          fullStat = t.reverseStm[stat]


          if t.primaryStats[fullStat] then
              primaryCount = primaryCount + 1
          end

          if fullStat == primaryStat then
              containsMyPrimary = true
          end

          if fullStat then
              res[fullStat] = (res[fullStat] or 0) + val
          end
      elseif line == "Prismatic Socket" then
          -- Add stats from socket
          if gemStat and gemStatVal then
              res[gemStat] = (res[gemStat] or 0) + gemStatVal
          end
      end
   end
   
   if containsMyPrimary and (primaryCount > 1) then      
      for k,v in pairs(res) do
         if t.primaryStats[k] and (k ~= primaryStat) then
            res[k] = nil
         end
         
      end
   end

   return res
end



t.registerScoreFactorFunction = function(f)
    t.scoreFactorFunction = f
end











